"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras.src.mixed_precision.dtype_policy import DTypePolicy
from keras.src.mixed_precision.dtype_policy import DTypePolicy as Policy
from keras.src.mixed_precision.dtype_policy import dtype_policy
from keras.src.mixed_precision.dtype_policy import dtype_policy as global_policy
from keras.src.mixed_precision.dtype_policy import set_dtype_policy
from keras.src.mixed_precision.dtype_policy import set_dtype_policy as set_global_policy
from keras.src.optimizers.loss_scale_optimizer import LossScaleOptimizer
